#!/usr/bin/env python3
from .base import Serializable, Searchable, Collectable
from . import fields


class Coordinates(Serializable):
    lat = fields.Field(float, none=False)
    lon = fields.Field(float, none=False)

    def __init__(self, lat=None, lon=None, **kwargs):
        super().__init__(lat=lat, lon=lon, **kwargs)

    def _serialize_instance(self, **kwargs):
        return [self.lat, self.lon]

    @classmethod
    def unserialize(cls, data):
        return cls(*data)

    def __eq__(self, other):
        return isinstance(other, Coordinates) and other.lat == self.lat and other.lon == self.lon

    def __repr__(self):
        return 'Coordinates(%.6f, %.6f)' % (self.lat, self.lon)


class AbstractLocation(Collectable):
    coords = fields.Model(Coordinates)

    def __init__(self, **kwargs):
        # magic, do not remove
        super().__init__(**kwargs)

    def __repr__(self):
        return 'AbstractLocation(%s)' % (repr(self.coords) if self.coords else '')

    def _near(self, other):
        return (self.coords is not None and other.coords is not None and
                abs(self.coords.lat - other.coords.lat) < 0.02 and
                abs(self.coords.lon - other.coords.lon) < 0.02)

    def _not_too_far(self, other):
        return ((self.coords is not None and other.coords is not None) and
                (abs(self.coords.lat - other.coords.lat) < 0.04 or
                 abs(self.coords.lon - other.coords.lon) < 0.04))

    def __eq__(self, other):
        return self.coords == other.coords

    def __ne__(self, other):
        return self.coords != other.coords


class Platform(AbstractLocation):
    stop = fields.Model('Stop', none=False)
    ifopt = fields.Field(str)
    name = fields.Field(str)
    full_name = fields.Field(str)

    def __init__(self, stop=None, name=None, full_name=None, **kwargs):
        super().__init__(stop=stop, name=name, full_name=full_name, **kwargs)

    def __repr__(self):
        return 'Platform(%s, %s, %s)' % (repr(self.stop), repr(self.name), repr(self.full_name))

    def __eq__(self, other):
        if not isinstance(other, Platform):
            return False

        by_id = self._same_by_id(other)
        if by_id is not None:
            return by_id

        if self.ifopt is not None and self.ifopt == other.ifopt:
            return True

        if self.stop != other.stop:
            return False

        if self.coords is not None and self.coords == other.coords:
            return True

        return None


class Location(AbstractLocation):
    country = fields.Field(str)
    city = fields.Field(str)
    name = fields.Field(str)
    near_stops = fields.Model('Stop.Results')

    def __init__(self, country=None, city=None, name=None, **kwargs):
        super().__init__(country=country, city=city, name=name, **kwargs)

    def __repr__(self):
        return '%s(%s, %s, %s)' % (self.__class__.__name__, repr(self.country), repr(self.city), repr(self.name))

    def __eq__(self, other):
        if not isinstance(other, Location):
            return False

        by_id = self._same_by_id(other)
        if by_id is not None:
            return by_id

        if (self.city is not None and self.city == other.city) or self._not_too_far(other):
            if self.name is not None and self.name == other.name:
                return True

        return None

    class Request(Searchable.Request):
        name = fields.Field(str)
        city = fields.Field(str)

        def __init__(self):
            super().__init__()
            self.name = None
            self.city = None


class Stop(Location):
    full_name = fields.Field(str)
    ifopt = fields.Field(str)
    uic = fields.Field(str)
    rides = fields.Model('Ride.Results')
    lines = fields.Model('Line.Results')

    def __init__(self, country=None, city=None, name=None, **kwargs):
        super().__init__(country=country, city=city, name=name, **kwargs)

    def _serialize_instance(self, **kwargs):
        if 'stops_had_results' not in kwargs:
            kwargs['stops_had_results'] = []

        if self.id not in kwargs['stops_had_results']:
            kwargs['stops_had_results'].append(self.id)
        else:
            kwargs['exclude'] = ['rides', 'lines']

        return super()._serialize_instance(**kwargs)

    def __repr__(self):
        return '<Stop %s>' % repr(self.full_name if self.full_name else (self.city, self.name))

    def __eq__(self, other):
        if not isinstance(other, Stop):
            return False

        by_id = self._same_by_id(other)
        if by_id is True:
            return True

        if (self.full_name is not None and
                self.full_name.replace(',', '') == other.full_name.replace(',', '')):
            return True

        if self.city is not None and other.city is not None:
            if self.city.split(' ')[0].lower() != other.city.split(' ')[0].lower():
                return False

            if self.city.startswith(other.city) and other.city.startswith(self.city):
                return self.name == other.name

        return None


class POI(Location):
    def __init__(self, country=None, city=None, name=None, **kwargs):
        super().__init__(country=country, city=city, name=name, **kwargs)

    @property
    def full_name(self):
        if self.city is None:
            return self.name
        else:
            return '%s, %s' % (self.city, self.name)

    def __eq__(self, other):
        if not isinstance(other, POI):
            return False

        return super().__eq__(other)


class Address(Location):
    street = fields.Field(str)
    number = fields.Field(str)

    def __init__(self, country=None, city=None, name=None, **kwargs):
        super().__init__(country=country, city=city, name=name, **kwargs)

    @property
    def full_name(self):
        if self.city is None:
            return self.name
        else:
            return '%s, %s' % (self.city, self.name)

    def __eq__(self, other):
        if not isinstance(other, Address):
            return False

        return super().__eq__(other)
